<?php
class WC_Stripe_Update {

	private static $updates = [ 
			'3.0.7' => 'update-3.0.7.php', 
			'3.1.0' => 'update-3.1.0.php' 
	];

	public static function init() {
		add_action ( 'init', [ __CLASS__, 'update' 
		] );
	}

	/**
	 * Performs an update on the plugin if required.
	 */
	public static function update() {
		// if option is not set, make the default version 3.0.6.
		$current_version = get_option ( 'stripe_wc_version', '3.0.6' );
		
		// if database version is less than plugin version, an update might be required.
		if (version_compare ( $current_version, wc_stripe ()->version (), '<' )) {
			foreach ( self::$updates as $version => $path ) {
				/*
				 * If the current version is less than the version in the loop, then perform upgrade.
				 */
				if (version_compare ( $current_version, $version, '<' )) {
					$file = wc_stripe ()->plugin_path () . 'includes/updates/' . $path;
					if (file_exists ( $file )) {
						include $file;
					}
					$current_version = $version;
					update_option ( 'stripe_wc_version', $current_version );
					add_action ( 'admin_notices', function () use ($current_version) {
						$message = sprintf ( __ ( 'Thank you for updating Stripe for WooCommerce to version %1$s.', 'woo-stripe-payment' ), $current_version );
						if (( $text = self::get_messages ( $current_version ) )) {
							$message .= ' ' . $text;
						}
						printf ( '<div class="notice notice-success is-dismissible"><p>%1$s</p></div>', $message );
					} );
				}
			}
			// save latest version.
			update_option ( 'stripe_wc_version', wc_stripe ()->version () );
		}
	}

	public static function get_messages($version) {
		$messages = [];
		return isset ( $messages[ $version ] ) ? $messages[ $version ] : false;
	}
}
WC_Stripe_Update::init ();